==> pips-1.01.tar.gz: pips-1.01/filters/mandel.c <==
/* PIPS version 1.01: Parallel Information Processing System 
   Copyright (C) 1994, 95, 96 Free Software Foundation, Inc.

This file is part of GNU PIPS.

GNU PIPS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU PIPS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/*
 * Name          : mandel.c
 * Author        : Frank Heimes
 * Institution   : Imperial College of Science, Technology, and Medicine
 * Written on    : Thu Oct 13 21:36:52 WET DST 1994
 * Modified on   : 
 *
 *  This file contains two functions:
 *
 *     pipsJobManager(), which is executed on the local host and
 *     pipsProcPar(),    which is executed on all hosts.
 *
 *  Both have their own description appended below.
 */

#include "pips.h"   /* PIPS system header file  */
#include "pvm3.h"   /* PVM system header file   */
#include <math.h>   /* Used: sqrt()             */

/*
 * +++ Set the minimum number of variables and parameters, 
 * the program expects on the command line
 */
int parameterNumber = 1;

/*
 * +++ Description of program for display on stderr by pipsPrintError()
 */
char *programName = "mandel";           /* Used in program description */
char *description[] = {
  "  mandel [-h] ncName var [height [width [rMi [rMa [cMi [cMa [iMa]]]]]]]",
  " ",
  "Where:",
  "  -h         Prints this help text.",
  "  ncName     Name of netCDF file to process, must be first parameter.",
  "  var        Name of variable to store image in",
  "  height     Window height in pixel",
  "  width      Window width in pixel",
  "  rMi        Minimum real value for left of window",
  "  rMa        Maximum real value for right of window",
  "  cMi        Minimum complex value for the bottom of window",
  "  cMa        Maximum complex value for the top of window",
  "  iMa        Maximum number of iterations after point is considered",
  "             convergent.",
  " ",
  "Description:",
  "  mandel opens 'ncName', spawns tasks on all processors in the",
  "  virtual parallel machine and ",
  "  generates a mandelbrot image, that is stored in 'variable'.",
  ""
}; /* end description */




/******************************************************************************
 *  P A R E N T   P R O C E S S  (on local host)
 ******************************************************************************
 *
 *   void pipsJobManager (ntask, tid, ncId, params)
 *
 *   int ntask        IS   Number of tasks available for parallel
 *                         data processing
 *   int tid []       IS   An array with task Id's of the tasks
 *   int ncId         IS   Id of opened netcdf file, returned by ncopen()
 *   char **params    IS   Tail of argv containing command line options
 *
 * Description   : 
 *
 *   Here, the data SEGMENTATION is done.
 *   Segments the data into pieces that can be processed in parallel,
 *   sends them to the tasks (that run pipsProcPar) and recollects it.
 *   If the number of data fragments (left) is less than the number of
 *   tasks, a competition is started that also guaranties that all lost
 *   data (crashed hosts, etc) is reprocessed.   
 */

#ifdef __STDC__

void pipsJobManager (int ntask, int tid[], int ncId, int parac, char **parav)

#else

pipsJobManager (ntask, tid, ncId, parac, parav)
    int ntask;
    int tid[];
    int ncId;
    int parac;
    char **parav;

#endif
{
/*
 * Variables for the job distribution
 */
  int done = 0;         /* Total number of processed jobs                    */
  int *jobId;           /* Array with currently processed jobs (1 per task)  */
  int *perform;         /* how many jobs each task has allready accomplished */
  int index;            /* Index to tid[], jobId[] and perform[]             */
  long i;               /* Index to tid[], jobId[] and perform[]             */
  int job = 0;          /* Current job Id (0 .. njobs-1)                     */
  int finishedJob;      /* Id of job that has just been finished             */
  int newJobs = 1;      /* Indicates that new jobs are still available       */
  int njobs;            /* Number of data slabs to process parallel          */

/*
 * +++ Netcdf variables and variables for data segmentation
 * start, edge: recordDim, rowDim, columnDim
 */
  long start[3];        /* Start of each datapacket                          */
  long edge [3];        /* Shape (size) of each datapacket                   */
  int fileType;         /* Value to use for file type attribute              */
  nc_type dataType;     /* Data type used in the result variable             */
  int var0Id;           /* Id of the first variable                          */
  long *result;         /* holds data that has been processed                */
  long rows;            /* Number of rows of data in one data fragment       */
  long firstLine;       /* Number of first line to compute                   */
  long images;          /* Number of images in current variable              */
  long height;          /* Height of image in variable 'varName' (dataId)    */
  long width;           /* Width of image in variable 'varName' (dataId)     */
  long maxIter;         /* Maximum number of iterations                      */
  nc_type resType;      /* Type of an existing variable for the results      */
  float rMin;           /* Minimum real value      */
  float rMax;           /* Maximum real value      */
  float cMin;           /* Minimum complex value   */
  float cMax;           /* Maximum complex value   */

/*
 * +++ Set the granularity to adjust the programs performance (see document.)
 */
  int granularity = 3;  /* The higher it is, the more jobs are generated     */


/******************************************************************************
 * +++  P R O C E S S   P A R A M E T E R S   A N D   P R E P A R A T I O N S 
 ******************************************************************************
 */

/*
 * +++ Evaluate the rest of command line parameters
 */
  index = 1;

  index++;
  height = 256;                    /* Default value */
  if (parac > index)
    height = atol (parav[index]);

  index++;
  width = 256;                    /* Default value */
  if (parac > index)
    width = atol (parav[index]);

  index++;
  rMin = -2.0;                    /* Default value */
  if (parac > index)
    rMin = atof (parav[index]);

  index++;
  rMax = 2.0;                    /* Default value */
  if (parac > index)
    rMax = atof (parav[index]);

  index++;
  cMin = -2.0;                    /* Default value */
  if (parac > index)
    cMin = atof (parav[index]);

  index++;
  cMax = 2.0;                    /* Default value */
  if (parac > index)
    cMax = atof (parav[index]);

  index++;
  maxIter = 255;                    /* Default value */
  if (parac > index)
    maxIter = atol (parav[index]);

/*
 * +++ Select type of outgoing data and fileType to create from it (later on)
 */
  dataType = NC_LONG;
  fileType = PipsPpmRaw;

/*
 * +++ Create the output variable (exit on failure)
 */
  resType = dataType;
  if ((var0Id = pipsVarDef (ncId, parav[0], &resType, &fileType,
                             &i, &height, &width)) < 0) {
    pipsPrintError (ncId, PipsInvalidVar, parav[0], PipsDontExit);
    return;
  } /* end if */

/*
 * Complain and exit, if the result variable exists and is unsuited for result
 */
  if (dataType != resType) {
    pipsPrintError (ncId, PipsVariableType, parav[0], PipsDontExit);
    return;
  } /* end if */

/*
 * +++ Define start and shape of data fragments (constant parts)
 */
  start [0] = 0;    edge  [0] = 1;
  start [2] = 0;    edge  [2] = width;

/*
 * +++ Determine the total size of data per variable
 */
  rows = height;

/*
 * Calculate the number of jobs (data fragments) and rows per job
 */
  rows = pipsDataPerJob (ntask,granularity,images,sqrt(rows),rows,&njobs);

/*
 * Reset the processed job array - currently no task is processing a job
 */
  jobId = pipsNewFlagArray (ntask, PipsNoJob);

/*
 * Reset the number of jobs accomplished by each task
 */
  perform = pipsNewFlagArray (ntask, 0);

/*
 * +++ Allocate memory, one buffer for both, loading and saving of data
 * Make sure the buffer size is a multiple of the larger data type
 */
  result = (long *) malloc (width * rows * sizeof (long));

/******************************************************************************
 *  D A T A   S E G M E N T A T I O N
 ******************************************************************************
 * Distribute jobs until all have been processed
 */
  while (done < njobs) {
    pipsAwaitChildData (tid,jobId,perform,&done, ntask, &finishedJob, &index);
/******************************************************************************
 * +++ R E C E I V E   A N D   S T O R E   P R O C E S S E D   D A T A 
 ******************************************************************************
 */
/*
 * Unpack and save data, if task was busy with a job
 */
    if (finishedJob != PipsNoJob) {
/*
 * +++ Set the row number, the data fragment starts at
 */
      start [1] = finishedJob * rows;
      edge  [1] = rows;
      pvm_upklong (result, edge[0]*edge[1]*edge[2], 1);
      ncvarput (ncId, var0Id, start, edge, (void *) result);
      fprintf (stderr, "Creating image, %d %% done.\n",(int)(100.0*done/njobs));
    } /* end if finishedJob */

/*
 * Keep sending, while there is still an unfinished job left
 */
    if (pipsJobAvailable (jobId, &job, ntask, index, newJobs)) {
/******************************************************************************
 * +++  L O A D   A N D   S E N D   P R O C E S S   D A T A 
 ******************************************************************************
 */
      pvm_pklong  (&width, 1, 1);
      pvm_pklong  (&height, 1, 1);
      firstLine = job * rows;
      pvm_pklong  (&firstLine, 1, 1);
      pvm_pklong  (&rows, 1, 1);
      pvm_pklong  (&maxIter, 1, 1);
      pvm_pkfloat (&rMin, 1, 1);
      pvm_pkfloat (&rMax, 1, 1);
      pvm_pkfloat (&cMin, 1, 1);
      pvm_pkfloat (&cMax, 1, 1);

/*
 * Send data packet to the idle child and tell it to commence work on it
 */
      pvm_send (tid[index], PipsCommence);
    } /* end if pipsJobAvailable */

/*
 * Select a job that hasn't been processed yet, if available
 * (for the next idle task)
 */
    pipsSelectNewJob (&newJobs, njobs, &job);
  } /* end while done < njobs */


/******************************************************************************
 *  E V A L U A T I O N   A N D   C L E A N U P
 ******************************************************************************
 */
/*
 * Report performance evaluation
 */
  pipsPrintPerformance (tid, perform, ntask, njobs);

/*
 * Free allocated memory
 */
  free (jobId);
  free (perform);
} /* pipsJobManager */







/******************************************************************************
 *  C H I L D   P R O C E S S  (on all hosts)
 ******************************************************************************
 *
 *   void pipsProcPar (void)
 *
 * Description   : 
 *
 *   Here the data PROCESSING is performed.
 *   Unpacks data received from the pipsJobManager (), processes it
 *   and packs the result data in again to be send back to the
 *   pipsJobManager. 
 */

#ifdef __STDC__

void pipsProcPar (void)

#else

pipsProcPar ()

#endif
{
/*
 * +++ Variables sent by the jobManager
 * NOTE: Declaring all variables 'static' increases speed dramatically
 */
  static long width;
  static long height;
  static long start;
  static long rows = -1;
  static long maxIter;
  static float rMin;
  static float rMax;
  static float cMin;
  static float cMax;

/*
 * +++ 'Backup' of above variables to keep track of changes in variables
 */
  static long currRows  = 0;

/*
 * +++ Variables for processing the data
 */
  static long ix, iy;                     /* loop indices */
  register double x, y;            /* pixel coords */
  register double ar, ai, a;       /* accums */
  register long ite;               /* number of iter until divergence */
  static unsigned long *result = NULL;

/*
 * +++ Define macros for better readability
 */
#define RESULT(X,Y) result [(Y)*width + (X)]

/******************************************************************************
 * R E C E I V E   P R O C E S S   D A T A 
 ******************************************************************************
 * +++ Unpack data that has arrived from the jobmanager
 */
  pvm_upklong (&width, 1, 1);
  pvm_upklong (&height, 1, 1);
  pvm_upklong (&start, 1, 1);
  pvm_upklong (&rows, 1, 1);
  pvm_upklong (&maxIter, 1, 1);
  pvm_upkfloat (&rMin, 1, 1);
  pvm_upkfloat (&rMax, 1, 1);
  pvm_upkfloat (&cMin, 1, 1);
  pvm_upkfloat (&cMax, 1, 1);

/*
 * +++ Preparations after change of parameters (data from different image)
 * Memory has to be reallocated and variables re-initialized
 */
  if (rows != currRows) {
    currRows = rows;

/*
 * +++ Allocate memory for variables
 */
    if (result)             /* Free allocated memory first */
      free (result);
    result = (unsigned long *) calloc (width * rows, sizeof (long));

  } /* end if */

/******************************************************************************
 * +++ C A L C U L A T I O N S   O N   D A T A 
 ******************************************************************************
 */
  rMax -= rMin;
  cMin -= cMax;
  for (iy = start; iy < start+rows; iy++) {
    y = (cMin * iy ) / height + cMax;
    for (ix = 0; ix < width; ix++) {
      x = (rMax * ix) / width + rMin;
      ar = x;
      ai = y;
      for (ite = 0; ite < maxIter; ite++) {
        a = (ar * ar) + (ai * ai);
        if (a > 4.0) break;
        a = ar * ar - ai * ai + x;
        ai = 2 * ai * ar + y;
        ar = a;
      }
      RESULT (ix, iy-start) = (maxIter-ite) % 256;
    }
  }

  pvm_initsend (PvmDataDefault);
/******************************************************************************
 * S E N D   R E S U L T   D A T A 
 ******************************************************************************
 * +++ Pack result data and send to jobmanager (sending done by main() )
 */
  pvm_pkulong (result, width * rows, 1);

} /* end pipsProcPar */


/* end mandel.c */

==> pips-1.02.tar.gz: pips-1.02/filters/mandel.c <==
/* PIPS version 1.01: Parallel Information Processing System 
   Copyright (C) 1994, 95, 96 Free Software Foundation, Inc.

This file is part of GNU PIPS.

GNU PIPS is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU PIPS is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU PIPS; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.
*/

/*
 * Name          : mandel.c
 * Author        : Frank Heimes
 * Institution   : Imperial College of Science, Technology, and Medicine
 * Written on    : Thu Oct 13 21:36:52 WET DST 1994
 * Modified on   : 
 *
 *  This file contains two functions:
 *
 *     pipsJobManager(), which is executed on the local host and
 *     pipsProcPar(),    which is executed on all hosts.
 *
 *  Both have their own description appended below.
 */

#include "pips.h"   /* PIPS system header file  */
#include "pvm3.h"   /* PVM system header file   */
#include <math.h>   /* Used: sqrt()             */

/*
 * +++ Set the minimum number of variables and parameters, 
 * the program expects on the command line
 */
int parameterNumber = 1;

/*
 * +++ Description of program for display on stderr by pipsPrintError()
 */
char *programName = "mandel";           /* Used in program description */
char *description[] = {
  "  mandel [-h] ncName var [height [width [rMi [rMa [cMi [cMa [iMa]]]]]]]",
  " ",
  "Where:",
  "  -h         Prints this help text.",
  "  ncName     Name of netCDF file to process, must be first parameter.",
  "  var        Name of variable to store image in",
  "  height     Window height in pixel",
  "  width      Window width in pixel",
  "  rMi        Minimum real value for left of window",
  "  rMa        Maximum real value for right of window",
  "  cMi        Minimum complex value for the bottom of window",
  "  cMa        Maximum complex value for the top of window",
  "  iMa        Maximum number of iterations after point is considered",
  "             convergent.",
  " ",
  "Description:",
  "  mandel opens 'ncName', spawns tasks on all processors in the",
  "  virtual parallel machine and ",
  "  generates a mandelbrot image, that is stored in 'variable'.",
  ""
}; /* end description */




/******************************************************************************
 *  P A R E N T   P R O C E S S  (on local host)
 ******************************************************************************
 *
 *   void pipsJobManager (ntask, tid, ncId, params)
 *
 *   int ntask        IS   Number of tasks available for parallel
 *                         data processing
 *   int tid []       IS   An array with task Id's of the tasks
 *   int ncId         IS   Id of opened netcdf file, returned by ncopen()
 *   char **params    IS   Tail of argv containing command line options
 *
 * Description   : 
 *
 *   Here, the data SEGMENTATION is done.
 *   Segments the data into pieces that can be processed in parallel,
 *   sends them to the tasks (that run pipsProcPar) and recollects it.
 *   If the number of data fragments (left) is less than the number of
 *   tasks, a competition is started that also guaranties that all lost
 *   data (crashed hosts, etc) is reprocessed.   
 */

#ifdef __STDC__

void pipsJobManager (int ntask, int tid[], int ncId, int parac, char **parav)

#else

pipsJobManager (ntask, tid, ncId, parac, parav)
    int ntask;
    int tid[];
    int ncId;
    int parac;
    char **parav;

#endif
{
/*
 * Variables for the job distribution
 */
  int done = 0;         /* Total number of processed jobs                    */
  int *jobId;           /* Array with currently processed jobs (1 per task)  */
  int *perform;         /* how many jobs each task has allready accomplished */
  int index;            /* Index to tid[], jobId[] and perform[]             */
  long i;               /* Index to tid[], jobId[] and perform[]             */
  int job = 0;          /* Current job Id (0 .. njobs-1)                     */
  int finishedJob;      /* Id of job that has just been finished             */
  int newJobs = 1;      /* Indicates that new jobs are still available       */
  int njobs;            /* Number of data slabs to process parallel          */

/*
 * +++ Netcdf variables and variables for data segmentation
 * start, edge: recordDim, rowDim, columnDim
 */
  long start[3];        /* Start of each datapacket                          */
  long edge [3];        /* Shape (size) of each datapacket                   */
  int fileType;         /* Value to use for file type attribute              */
  nc_type dataType;     /* Data type used in the result variable             */
  int var0Id;           /* Id of the first variable                          */
  long *result;         /* holds data that has been processed                */
  long rows;            /* Number of rows of data in one data fragment       */
  long firstLine;       /* Number of first line to compute                   */
  long images;          /* Number of images in current variable              */
  long height;          /* Height of image in variable 'varName' (dataId)    */
  long width;           /* Width of image in variable 'varName' (dataId)     */
  long maxIter;         /* Maximum number of iterations                      */
  nc_type resType;      /* Type of an existing variable for the results      */
  float rMin;           /* Minimum real value      */
  float rMax;           /* Maximum real value      */
  float cMin;           /* Minimum complex value   */
  float cMax;           /* Maximum complex value   */

/*
 * +++ Set the granularity to adjust the programs performance (see document.)
 */
  int granularity = 3;  /* The higher it is, the more jobs are generated     */


/******************************************************************************
 * +++  P R O C E S S   P A R A M E T E R S   A N D   P R E P A R A T I O N S 
 ******************************************************************************
 */

/*
 * +++ Evaluate the rest of command line parameters
 */
  index = 1;

  index++;
  height = 256;                    /* Default value */
  if (parac > index)
    height = atol (parav[index]);

  index++;
  width = 256;                    /* Default value */
  if (parac > index)
    width = atol (parav[index]);

  index++;
  rMin = -2.0;                    /* Default value */
  if (parac > index)
    rMin = atof (parav[index]);

  index++;
  rMax = 2.0;                    /* Default value */
  if (parac > index)
    rMax = atof (parav[index]);

  index++;
  cMin = -2.0;                    /* Default value */
  if (parac > index)
    cMin = atof (parav[index]);

  index++;
  cMax = 2.0;                    /* Default value */
  if (parac > index)
    cMax = atof (parav[index]);

  index++;
  maxIter = 255;                    /* Default value */
  if (parac > index)
    maxIter = atol (parav[index]);

/*
 * +++ Select type of outgoing data and fileType to create from it (later on)
 */
  dataType = NC_LONG;
  fileType = PipsPpmRaw;

/*
 * +++ Create the output variable (exit on failure)
 */
  resType = dataType;
  if ((var0Id = pipsVarDef (ncId, parav[0], &resType, &fileType,
                             &i, &height, &width)) < 0) {
    pipsPrintError (ncId, PipsInvalidVar, parav[0], PipsDontExit);
    return;
  } /* end if */

/*
 * Complain and exit, if the result variable exists and is unsuited for result
 */
  if (dataType != resType) {
    pipsPrintError (ncId, PipsVariableType, parav[0], PipsDontExit);
    return;
  } /* end if */

/*
 * +++ Define start and shape of data fragments (constant parts)
 */
  start [0] = 0;    edge  [0] = 1;
  start [2] = 0;    edge  [2] = width;

/*
 * +++ Determine the total size of data per variable
 */
  rows = height;

/*
 * Calculate the number of jobs (data fragments) and rows per job
 */
  rows = pipsDataPerJob (ntask,granularity,images,sqrt(rows),rows,&njobs);

/*
 * Reset the processed job array - currently no task is processing a job
 */
  jobId = pipsNewFlagArray (ntask, PipsNoJob);

/*
 * Reset the number of jobs accomplished by each task
 */
  perform = pipsNewFlagArray (ntask, 0);

/*
 * +++ Allocate memory, one buffer for both, loading and saving of data
 * Make sure the buffer size is a multiple of the larger data type
 */
  result = (long *) malloc (width * rows * sizeof (long));

/******************************************************************************
 *  D A T A   S E G M E N T A T I O N
 ******************************************************************************
 * Distribute jobs until all have been processed
 */
  while (done < njobs) {
    pipsAwaitChildData (tid,jobId,perform,&done, ntask, &finishedJob, &index);
/******************************************************************************
 * +++ R E C E I V E   A N D   S T O R E   P R O C E S S E D   D A T A 
 ******************************************************************************
 */
/*
 * Unpack and save data, if task was busy with a job
 */
    if (finishedJob != PipsNoJob) {
/*
 * +++ Set the row number, the data fragment starts at
 */
      start [1] = finishedJob * rows;
      edge  [1] = rows;
      pvm_upklong (result, edge[0]*edge[1]*edge[2], 1);
      ncvarput (ncId, var0Id, start, edge, (void *) result);
      fprintf (stderr, "Creating image, %d %% done.\n",(int)(100.0*done/njobs));
    } /* end if finishedJob */

/*
 * Keep sending, while there is still an unfinished job left
 */
    if (pipsJobAvailable (jobId, &job, ntask, index, newJobs)) {
/******************************************************************************
 * +++  L O A D   A N D   S E N D   P R O C E S S   D A T A 
 ******************************************************************************
 */
      pvm_pklong  (&width, 1, 1);
      pvm_pklong  (&height, 1, 1);
      firstLine = job * rows;
      pvm_pklong  (&firstLine, 1, 1);
      pvm_pklong  (&rows, 1, 1);
      pvm_pklong  (&maxIter, 1, 1);
      pvm_pkfloat (&rMin, 1, 1);
      pvm_pkfloat (&rMax, 1, 1);
      pvm_pkfloat (&cMin, 1, 1);
      pvm_pkfloat (&cMax, 1, 1);

/*
 * Send data packet to the idle child and tell it to commence work on it
 */
      pvm_send (tid[index], PipsCommence);
    } /* end if pipsJobAvailable */

/*
 * Select a job that hasn't been processed yet, if available
 * (for the next idle task)
 */
    pipsSelectNewJob (&newJobs, njobs, &job);
  } /* end while done < njobs */


/******************************************************************************
 *  E V A L U A T I O N   A N D   C L E A N U P
 ******************************************************************************
 */
/*
 * Report performance evaluation
 */
  pipsPrintPerformance (tid, perform, ntask, njobs);

/*
 * Free allocated memory
 */
  free (jobId);
  free (perform);
} /* pipsJobManager */







/******************************************************************************
 *  C H I L D   P R O C E S S  (on all hosts)
 ******************************************************************************
 *
 *   void pipsProcPar (void)
 *
 * Description   : 
 *
 *   Here the data PROCESSING is performed.
 *   Unpacks data received from the pipsJobManager (), processes it
 *   and packs the result data in again to be send back to the
 *   pipsJobManager. 
 */

#ifdef __STDC__

void pipsProcPar (void)

#else

pipsProcPar ()

#endif
{
/*
 * +++ Variables sent by the jobManager
 * NOTE: Declaring all variables 'static' increases speed dramatically
 */
  static long width;
  static long height;
  static long start;
  static long rows = -1;
  static long maxIter;
  static float rMin;
  static float rMax;
  static float cMin;
  static float cMax;

/*
 * +++ 'Backup' of above variables to keep track of changes in variables
 */
  static long currRows  = 0;

/*
 * +++ Variables for processing the data
 */
  static long ix, iy;                     /* loop indices */
  register double x, y;            /* pixel coords */
  register double ar, ai, a;       /* accums */
  register long ite;               /* number of iter until divergence */
  static unsigned long *result = NULL;

/*
 * +++ Define macros for better readability
 */
#define RESULT(X,Y) result [(Y)*width + (X)]

/******************************************************************************
 * R E C E I V E   P R O C E S S   D A T A 
 ******************************************************************************
 * +++ Unpack data that has arrived from the jobmanager
 */
  pvm_upklong (&width, 1, 1);
  pvm_upklong (&height, 1, 1);
  pvm_upklong (&start, 1, 1);
  pvm_upklong (&rows, 1, 1);
  pvm_upklong (&maxIter, 1, 1);
  pvm_upkfloat (&rMin, 1, 1);
  pvm_upkfloat (&rMax, 1, 1);
  pvm_upkfloat (&cMin, 1, 1);
  pvm_upkfloat (&cMax, 1, 1);

/*
 * +++ Preparations after change of parameters (data from different image)
 * Memory has to be reallocated and variables re-initialized
 */
  if (rows != currRows) {
    currRows = rows;

/*
 * +++ Allocate memory for variables
 */
    if (result)             /* Free allocated memory first */
      free (result);
    result = (unsigned long *) calloc (width * rows, sizeof (long));

  } /* end if */

/******************************************************************************
 * +++ C A L C U L A T I O N S   O N   D A T A 
 ******************************************************************************
 */
  rMax -= rMin;
  cMin -= cMax;
  for (iy = start; iy < start+rows; iy++) {
    y = (cMin * iy ) / height + cMax;
    for (ix = 0; ix < width; ix++) {
      x = (rMax * ix) / width + rMin;
      ar = x;
      ai = y;
      for (ite = 0; ite < maxIter; ite++) {
        a = (ar * ar) + (ai * ai);
        if (a > 4.0) break;
        a = ar * ar - ai * ai + x;
        ai = 2 * ai * ar + y;
        ar = a;
      }
      RESULT (ix, iy-start) = (maxIter-ite) % 256;
    }
  }

  pvm_initsend (PvmDataDefault);
/******************************************************************************
 * S E N D   R E S U L T   D A T A 
 ******************************************************************************
 * +++ Pack result data and send to jobmanager (sending done by main() )
 */
  pvm_pkulong (result, width * rows, 1);

} /* end pipsProcPar */


/* end mandel.c */
